{-
This module contains the inference functions to infer both the Agt specification 
and the haskell function equations from the circular specification.

In the future, this two inference processes are to be separated, since one of them
is target language independent and the other is target language dependent.
-}

module Ocaml2AGT where


import AgtTypes
import FunctionTypes

import List
import Char

import Language.Haskell.Syntax
import Language.Haskell.Parser

--import Debug.Trace


-- function that, given the list of the module's HsDecls, splits them in its data declarations,
-- its type signatures, its pattern bindings and its function bindings
split3 :: (Num a) =>
          a
          -> ([HsDecl], [HsDecl], [HsDecl], [HsDecl])
          -> [HsDecl]
          -> ([HsDecl], [HsDecl], [HsDecl], [HsDecl])
split3 _ t [] = t
split3 x (dt,ts,pb,ft) (h:t) = 
	let (dt',ts',pb',ft') = split3 x (dt,ts,pb,ft) t
	in case h of 
		(HsDataDecl _ _ _ _ _ _)     -> (h:dt',ts',pb',ft')
		(HsTypeSig  _ _ _)           -> (dt',h:ts',pb',ft')
		(HsPatBind  _ _ _ _)         -> (dt',ts',h:pb',ft')
		(HsFunBind [])	             -> (dt',ts',pb',ft')
		(HsFunBind  
		   ((HsMatch _ _ _ _ l):ls)) -> let (dt'',ts'',pb'',ft'') = split3 x ([],[],[],[]) l
		   				in if (x==0) 
						   then split3 1 (dt''++dt', ts''++ts', pb', h:(ft''++ft')) [HsFunBind ls]
						   else split3 x (dt''++dt', ts''++ts', pb', ft''++ft')     [HsFunBind ls]
		_                            -> (dt',ts',pb',ft')


{-
  function that inferes the agt specification and the function 
  equations from the Haskell module tree;

  the non-treated functions and pattern bindings are also 
  (non-altered) computed, so that they can be copied into the
  strict module
-}

ocaml2Agt_Functions :: HsModule 
	    -> ((Agt,Functions),[HsDecl],[HsDecl])

ocaml2Agt_Functions (HsModule _ m _ _ ldecls)  = let 
	l@(datadecls,typesigs,patbinds,funbinds) = split3 0 ([],[],[],[]) ldecls
	nonterminals                    = NT $ getNonTerminals datadecls
	productions                     = getProductions       datadecls
	(treated_patbinds, ntreated_patbinds ) = getSliceItems patbinds []
	(treated_functions,ntreated_functions) = hsmatches2equations (concatMap (\(HsFunBind l) -> l) funbinds)
								     treated_patbinds 
								     []
	(equations,functions)           = unzip treated_functions
	attributes			= nub   $ getAttributes        equations datadecls
					 in ((Agt nonterminals
						 productions
						 attributes
						 equations,

					      functions
					     ),
                                             map (HsFunBind.(:[])) ntreated_functions,
					     ntreated_patbinds)
	where
		getNonTerminals  [] = []
		getNonTerminals  ((HsDataDecl _
					      _
					      (HsIdent nt)
					      _
					      _
					      _):t) = nt: getNonTerminals t 
		getProductions   [] = []
		getProductions   (h:t)
				    = getProductionsAux h ++ getProductions t 
			where 
			      getProductionsAux (HsDataDecl _
							    _
							    _
							    _
							    []
							    _) = []
			      getProductionsAux (HsDataDecl s@_
							    v@_
							    c@(HsIdent nt)
							    w@_
							    ((HsConDecl _
									(HsIdent cons)
									lsHsUnBangedTy):t)
							    z@_
							    ) = (Prod 
								   cons 
								   nt 
							           (nt:(analyze lsHsUnBangedTy))):
								getProductionsAux (HsDataDecl s v c w t z)
		analyze          [] = []
		analyze ((HsUnBangedTy
				(HsTyTuple l)):t)            = analyze ((map HsUnBangedTy l)++t)
		analyze ((HsUnBangedTy 
				(HsTyCon 
				   (UnQual 
					(HsIdent name)))):t) = analyzeAux name : analyze t
		analyze ((HsBangedTy
				(HsTyTuple l)):t)            = analyze ((map HsBangedTy l)++t)
		analyze ((HsBangedTy 
				(HsTyCon 
				   (UnQual 
					(HsIdent name)))):t) = analyzeAux name : analyze t
		analyze x = error $ "undefined error" ++ show x
		analyzeAux "Int"    = "INT"
		analyzeAux "Char"   = "CHAR"
		analyzeAux "Bool"   = "BOOL"
		analyzeAux "String" = "STR"
		analyzeAux x        = x
		{-...-}
		getAttributes    [] _ = []
		getAttributes    ((Eq _
				     cons
				     ((_,i,name):_)):t) datadecls
					= let attributeType = if i == 0 
							      then SYN 
							      else INH
					      nt            = getConstructorDataType cons i datadecls
					  in  (Attr name
						    attributeType
						    nt): getAttributes t datadecls
		getConstructorDataType cons 0 ((HsDataDecl _
							   _
							   (HsIdent datatype)
							   _
							   lHsConDecl
							   _):t) = if (isDefinedIn cons lHsConDecl)
								   then datatype
								   else getConstructorDataType cons 0 t
		getConstructorDataType cons i ((HsDataDecl _
							   _
							   (HsIdent datatype)
							   _
							   lHsConDecl
							   _):t) = let r = isDefinedIn' cons i lHsConDecl
								   in
								   if r == Nothing
								   then getConstructorDataType cons i t
								   else (\(Just x) -> x) r
		isDefinedIn  _    [] = False
		isDefinedIn  cons ((HsConDecl _
					      (HsIdent cons')
					      _):t) | cons == cons' = True
						    | otherwise     = isDefinedIn cons t
		isDefinedIn' cons _ [] = Nothing
		isDefinedIn' cons i ((HsConDecl x
						(HsIdent cons')
						[HsUnBangedTy (HsTyTuple l)]):t)
							 = isDefinedIn' cons i ((HsConDecl x
											   (HsIdent cons')
											   (map HsUnBangedTy l)):t)
		isDefinedIn' cons i ((HsConDecl x
						(HsIdent cons')
						[HsBangedTy (HsTyTuple l)]):t)
							 = isDefinedIn' cons i ((HsConDecl x
											   (HsIdent cons')
											   (map HsBangedTy l)):t) 
		isDefinedIn' cons i ((HsConDecl _
					       (HsIdent cons')
					       lHsTy):t) | cons == cons' = case (lHsTy !! (i-1)) of
											(HsUnBangedTy
												(HsTyCon
													(UnQual
														(HsIdent datatype)))) -> Just datatype
											(HsBangedTy 
												(HsTyCon
													(UnQual
														(HsIdent datatype)))) -> Just datatype
											x -> error $ "isDefinedIn' " ++ (show x)
												
						         | otherwise     = isDefinedIn' cons i t
		getSliceItems [] l                  = ([],l)
		getSliceItems (h@(HsPatBind _
					   (HsPVar (HsIdent "slice"))
					   (HsUnGuardedRhs (HsList sliceList))
					   _) :t) l = 
						let sliceItems = map (\ (HsTuple ((HsLit (HsString fname)):[HsList fargs])) 
										-> (fname,map (\ (HsLit (HsString arg)) -> arg) fargs)) sliceList
						in (sliceItems, l++t)
		getSliceItems (h:t) l               = let (slice,nt_patbinds) = getSliceItems t l
						      in  (slice, h:nt_patbinds)             
		getFunctionNames [] = []
		getFunctionNames ((HsFunBind ((HsMatch _
						      (HsIdent fname)
						      _
						      _
						      _):_)):t)= fname : getFunctionNames t 


-- 
hsmatches2equations :: [HsMatch]
		    -> [(FunctionName,[FunctionArgument])] -- slice items
		    -> [HsMatch]                           -- acumulator to keep record of the non treated cases
		    -> ([(Equation,Function)], [HsMatch])
hsmatches2equations [] 
		    _
		    ac = ([],ac) 
hsmatches2equations (m@(HsMatch _ 
			        (HsIdent fname)  
			        [HsPTuple ((HsPApp (UnQual (HsIdent id)) varnames):_)] 
			        _ 
			        patbinds):t)
		     slice 
		     ac
	= let sliceArgs = sliceArguments fname slice
	      patbinds' = filterPatBinds patbinds sliceArgs
	      l = patbinds2attrs patbinds' (concatMap aux varnames)
	      aux (HsPVar (HsIdent x)) = [x]
	      aux (HsPTuple l)         = concatMap aux l 
              h = if l == []
	          then (ls,m:ac') 
		  else ((map (\((eqType,attr),(occrs,fname,args)) 
				-> (Eq eqType 
				       id 
				       attr,
				    Function (id ++ "_" ++ occrs)
					     fname
					     (map ((id++"_")++) args))) l) ++ ls, ac')
	      (ls,ac') = hsmatches2equations t slice ac
	  in h

hsmatches2equations (m@(HsMatch x 
			        fname
			        ((HsPParen ((HsPApp (UnQual id) varnames))):y)
			        z 
			        patbinds):t) 
		     slice
		     ac 
	= hsmatches2equations ((HsMatch x
				        fname
					((HsPApp (UnQual id) varnames):y)
					z
					patbinds):t)
				slice
				ac

hsmatches2equations (m@(HsMatch _ 
			        (HsIdent fname) 
			        (((HsPApp (UnQual (HsIdent id)) varnames)):_)
			        _ 
			        patbinds):t) 
		     slice
		     ac 
	= let sliceArgs = sliceArguments fname slice
	      patbinds' = filterPatBinds patbinds sliceArgs
	      l = patbinds2attrs patbinds' (concatMap aux varnames)
	      aux (HsPVar (HsIdent x)) = [x]
	      aux (HsPTuple l)         = concatMap aux l 
              h = if l == []
	          then (ls,m:ac')
		  else ((map (\((eqType,attr),(occrs,fname,args)) 
				-> (Eq eqType 
				       id 
				       attr,
				    Function (id ++ "_" ++ occrs)
					     fname
					     (map ((id++"_")++) args))) l) ++ ls, ac')
	      (ls,ac') = hsmatches2equations t slice ac
	  in h

hsmatches2equations (h:t) slice ac =
		let (t',ac') = hsmatches2equations t slice ac
		in  (t',h:ac')
					 

{-

This functions filters the "interesting" patterns, whether they arise from
their shape or from slicing specifications

-}

filterPatBinds ::    [HsDecl]
                  -> [[FunctionArgument]]
                  -> [HsDecl]
filterPatBinds []
	       _  = []
filterPatBinds (h@(HsPatBind _
			     x
			     _
			     _):t)
		sliceArgs = case x of
				(HsPVar (HsIdent attr)) -> if (sliceArgs == [] || attr `elem` (head sliceArgs) || isInherited)
							   then h : (filterPatBinds t sliceArgs)
							   else      filterPatBinds t sliceArgs
						where isInherited = (\(x,y,z) -> y) (eficientPartition isDigit [] attr) /= 0
			        _                       -> filterPatBinds t sliceArgs


{-

Given a function name and the list of items to slice, this function
computes the list of the function's arguments to slice;

If the function is not to slice (this corresponds to stating the interest
in obtaining the full function's slice, or not to slice at all), the arguments 
to slice will be represented by the empty list;

If the function is to slice, but no slicing arguments are given, like in:

	slice = [("repmin",[])]

this case's slicing arguments will be represented as:

	[[]]

-}

sliceArguments :: FunctionName 
	       -> [(FunctionName, [FunctionArgument])]
	       -> [[FunctionArgument]]
sliceArguments _ [] = []
sliceArguments f1 ((f2,args):t) | f1 == f2  = [args]
				| otherwise = sliceArguments f1 t



{- 
   this function gets the list of the module functions' pattern bindings, and calculates,
   using the list of variable names and the list of functions' slice items, for each binding, 
   the corresponding pair 
				(( agt equation type, 
				   attrs ),
				 ( function name, 
				   function definition, 
				   function args)
				) 
   , if any
-}

patbinds2attrs :: [HsDecl] 
	       -> [String]
	       -> [((EquationType,Attrs),(String,String,[String]))]
patbinds2attrs [] 
	       _ = []


-- attr1 = (expr)
patbinds2attrs ((HsPatBind x1
		           attr
			   (HsUnGuardedRhs (HsParen expr))
			   x2):t)
		varnames
	= patbinds2attrs ((HsPatBind x1 attr (HsUnGuardedRhs expr) x2):t) varnames


-- attr1 = literal
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsLit lit))
			  _):t) 
		varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      eq             = ( (SynRule,
		 		  [a1]
				 ),

	        		(show (i1+1)++"_"++str1,
	        		 aux lit,
				 []
				) 
	      		       )
	  in eq : patbinds2attrs t varnames 

-- attr1 = []
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsList []))
			   _):t) 
		varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      eq             = ( (SynRule,
				  [a1]),

				 (show (i1+1)++"_"++str1,
				  "[]",
				  []
				  ))
	in eq : patbinds2attrs t varnames


-- attr1 = attr2
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsVar attr2))
			  _):t) 
		varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      eq = ( (CopyRule,
		      [a1,a2]
		     ),

	             (show (i1+1)++"_"++str1,
	              "",
		      [show i2 ++ "_" ++ str2]) 
	           )
	  in eq : patbinds2attrs t varnames 


-- attr1 = Constructor
-- where Constructor has 0 arguments
-- falta testar s o construtor est definido no mdulo!
-- s estiver, isto est correcto,seno,  s tirar o "C_" e o "_1"
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsCon cons))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	  in ( (SynRule,
	        [a1]
               ),

	      (show (i1+1)++"_"++str1,
	      "C_"++getVarName cons++"_1",
	      [])
	     )
		:patbinds2attrs t varnames


-- attr1 = Constructor (expr1)
patbinds2attrs ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsCon cons) (HsParen expr1)))
			   y):t) varnames
	= patbinds2attrs ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsCon cons) expr1))
			             y):t) varnames


-- attr1 = Constructor literal
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsCon cons)
						  (HsLit lit)))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	  in ( (SynRule,
	       [a1]
	       ),
	       
		(show (i1+1) ++ "_" ++ str1,
	        "C_"++getVarName cons++"_1 " ++ aux lit,
		[])
	     )
		:patbinds2attrs t varnames


-- attr1 = Constructor attr2
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsCon cons)
						  (HsVar attr2)))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	  in ( (SynRule,
	       [a1,a2]
	       ),
	       
		(show (i1+1) ++ "_" ++ str1,
	        "C_"++getVarName cons++"_1",
		[show i2 ++ "_" ++ str2])
	     )
		:patbinds2attrs t varnames


-- attr1 = Constructor (expr2) expr3
patbinds2attrs ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							 (HsParen expr2))
						  expr3))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							           expr2)
							    expr3))
				     x):t) varnames


-- attr1 = Constructor expr2 (expr3)
patbinds2attrs ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							 expr2)
						  (HsParen expr3)))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							           expr2)
							    expr3))
				     x):t) varnames


-- attr1 = Constructor (attr1, attr2)
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsCon cons)
						 (HsTuple [expr2, expr3])))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2attrs t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of
					HsVar attr3 -> -- Constructor (attr2, attr3) [No caso de querermos gerar ML, a funcao a gerar ser Cons(attr1,attr2)]
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  						in ( (SynRule,
								[a1,a2,a3]
	       						      ),

							      (show (i1+1) ++ "_" ++ str1,
		 					       "C_" ++ aux'' cons ++ "_1",
		 					       [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
								)
	     						   ) : r


-- attr1 = Constructor/special_function attr2              attr3
-- attr1 = Constructor/special_function literal            attr2
-- attr1 = Constructor/special_function [function literal] attr2
-- attr1 = Constructor/special_function attr2              literal
-- attr1 = Constructor/special_function [Constructor attr2] attr3
-- attr1 = Constructor/special_function [Constructor attr2 attr3] attr4
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsApp (HsCon cons)
							 expr2)
						  expr3))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
              r              = patbinds2attrs t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of
					HsVar attr3 -> -- attr1 = Constructor attr2 atrr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  						in ( (SynRule,
								[a1,a2,a3]
	       						      ),

							      (show (i1+1) ++ "_" ++ str1,
		 					       "C_" ++ aux'' cons ++ "_1",
		 					       [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
								)
	     						   ) : r
					HsLit lit   -> -- attr1 = Constructor attr2 literal
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
							in ( (SynRule,
								[a1,a2]
	       						     ),

							    (show (i1+1) ++ "_"++ str1,
							     "(\\x -> " ++ "C_" ++ aux'' cons ++ "_1 x " ++ aux lit ++ ")",
							     [show i2 ++ "_" ++ str2] 
							    )
	     						   ) : r
		HsLit lit  -> case expr3 of
					HsVar attr2 -> -- attr1 = Constructor literal attr2
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	  						in ( (SynRule,
								[a1,a2]
	       						     ),

	       						     (show (i1+1) ++ "_"++ str1,
							      "C_" ++ aux'' cons ++ "_1 " ++ aux lit,
							      [show i2 ++ "_" ++ str2]
							     )
	     						   ) : r
		HsApp (HsVar f)
		      (HsLit lit) -> case expr3 of
					HsVar attr2 -> -- attr1 = Constructor [function literal] attr2
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	  						in ( (SynRule,
								[a1,a2]
	       						     ),

							     (show (i1+1) ++ "_"++ str1,
		 					      "C_" ++ aux'' cons ++ "_1 (" ++ FunctionTypes.space (getVarName f) ++  aux lit ++ ")",
		 					      [show i2 ++ "_" ++ str2]
							     )
	     						   ) : r
		HsApp (HsCon (UnQual (HsIdent constructor)))
		      (HsVar attr2)
				-> case expr3 of
					HsVar attr3 -> -- attr1 = Constructor/special_function [Constructor attr2] attr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  						in ( (SynRule,
	       							[a1,a2,a3]
							     ),

							    (show (i1+1) ++ "_" ++ str1,
		 					     "(\\x y -> (" ++ aux'' cons ++ ") (C_" ++ constructor ++ "_1  x) y)",
							     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
							   )
	     						   ) : r
		HsApp (HsApp (HsCon (UnQual (HsIdent constructor)))
			     (HsVar attr2))
		      (HsVar attr3)
				-> case expr3 of 
					HsVar attr4 -> -- attr1 = Constructor/special_function [Constructor attr2 attr3] attr4
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  						in ( (SynRule,
	       							[a1,a2,a3,a4]
							     ),

							     (show (i1+1) ++ "_" ++ str1,
		 					      "(\\x y z -> (" ++ aux'' cons ++ ") (C_" ++ constructor ++ "_1  x y) z)",
							      [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
							     )
	    	 					   ) : r


-- attr1 = Constructor (attr2, attr3, attr4)
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsCon cons)
						 (HsTuple [expr2, expr3, expr4])))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2attrs t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of
					HsVar attr3 -> case expr4 of
								HsVar attr4 -> -- Constructor (attr2, attr3, attr4)
									let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  								in ( (SynRule,
										[a1,a2,a3,a4]
	       						      			),

							      		     (show (i1+1) ++ "_" ++ str1,
		 					       			"C_" ++ aux'' cons ++ "_1",
		 					       			[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
									     )
	     						   		    ) : r


-- attr1 = Constructor (expr2) expr3 expr4
patbinds2attrs ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								(HsParen expr2))
							 expr3) 
						  expr4))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons)
									  expr2)
								   expr3)
						      expr4))
				     x):t) varnames


-- attr1 = Constructor expr2 (expr3) expr4
patbinds2attrs ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								expr2)
							 (HsParen expr3)) 
						  expr4))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons)
									  expr2)
								   expr3)
						      expr4))
				     x):t) varnames


-- attr1 = Constructor expr2 expr3 (expr4)
patbinds2attrs ((HsPatBind b
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								expr2)
							 expr3) 
						  (HsParen expr4)))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons)
									  expr2)
								   expr3)
						      expr4))
				     x):t) varnames


-- attr1 = Constructor/special_function attr2 attr3 attr4
-- attr1 = Constructor/special_function literal attr3 attr4
-- attr1 = Constructor/special_function [function literal] attr3 attr4
-- attr1 = Constructor/special_function [function attr2] attr3 attr4, 
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsCon cons) 
								expr2)
							 expr3) 
						  expr4))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2attrs t varnames
	  in case expr2 of 
			HsVar attr2 -> case expr3 of
						HsVar attr3 -> case expr4 of
									HsVar attr4 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
											   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
											   a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
										       in  ( (SynRule,
												[a1,a2,a3,a4]
	       										      ),

											     (show (i1+1) ++ "_" ++ str1,
		 										"C_" ++ aux'' cons ++ "_1",
		 										[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
											     )
	      										    ) : r
			HsLit lit   -> case expr3 of
						HsVar attr2 -> case expr4 of
									HsVar attr3 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      										   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
										       in ( (SynRule,
												[a1,a2,a3]
	       										    ),

											    (show (i1+1) ++ "_"++ str1,
		 										"C_" ++ aux'' cons ++ "_1 " ++ aux lit,
		 										[show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
											    )
	     										  ) : r
			HsApp (HsVar f) (HsLit lit)
				    -> case expr3 of 
						HsVar attr2 -> case expr4 of
									HsVar attr3 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      										   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
										       in  ( (SynRule,
												[a1,a2,a3]
	       										     ),

											     (show (i1+1) ++ "_"++ str1,
		 									     "C_" ++ aux'' cons ++ "_1 (" ++ FunctionTypes.space (getVarName f) ++  aux lit ++ ")",
		 									     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
											     )
	     										    ) : r
			HsApp (HsVar f) (HsVar attr2)
				   -> case expr3 of
						HsVar attr3 -> case expr4 of
									HsVar attr4 -> let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      									           a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      									           a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
								 		       in ( (SynRule,
												[a1,a2,a3,a4]
	       										    ),

											    (show (i1+1) ++ "_"++ str1,
		 									     "(\\x y z -> " ++ "C_" ++ aux'' cons ++ "_1 (" ++ FunctionTypes.space (getVarName f) ++  "x) y z)",
		 									     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
											    )
	     										  ) : r


-- attr1 = function (expr2)
patbinds2attrs ((HsPatBind x
			  (HsPVar attr1)
			  (HsUnGuardedRhs (HsApp (HsVar f)
						 (HsParen expr2)))
			  y):t) varnames
	= patbinds2attrs ((HsPatBind x
					(HsPVar attr1)
					(HsUnGuardedRhs (HsApp (HsVar f)
							       expr2))
			 		y):t) varnames


-- attr1 = function attr2
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsVar (UnQual (HsIdent f)))
						 (HsVar (UnQual (HsIdent attr2)))))
			  _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
	  in ( (SynRule,
		     [a1,a2]
		    ),
		
		    (show (i1+1) ++ "_" ++ str1,
		     f,
		     [show i2 ++ "_" ++ str2]
		    )
		  )
		:patbinds2attrs t varnames


-- attr1 = function (expr2) expr3
patbinds2attrs ((HsPatBind x
			  (HsPVar attr1)
			  (HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
							(HsParen expr2))
						 expr3))
			  y):t) varnames
	= patbinds2attrs ((HsPatBind x
			  	(HsPVar attr1)
			  	(HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
						              expr2)
						       expr3))
			  y):t) varnames


-- attr1 = function expr2 (expr3)
patbinds2attrs ((HsPatBind x
			  (HsPVar attr1)
			  (HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
							expr2)
						 (HsParen expr3)))
			  y):t) varnames
	= patbinds2attrs ((HsPatBind x
			  	(HsPVar attr1)
			  	(HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual f))
						              expr2)
						       expr3))
			  y):t) varnames


-- attr1 = function attr2 atrr3
-- attr1 = function Constructor attr2
-- attr1 = function [Constructor attr2] attr3
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							expr2)
						 expr3))
			  _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r = patbinds2attrs t varnames
	  in case expr2 of
		HsVar attr2 -> case expr3 of 
					HsVar attr3 -> -- attr1 = function  attr2 attr3
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  					       in ( (SynRule,
		     						[a1,a2,a3]
		    					    ),

		    					    (show (i1+1) ++ "_" ++ str1,
		     					     f,
		     					     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
		    					    )
		  					) : r 
		HsCon cons -> case expr3 of 
					HsVar attr2 -> -- attr1 = function Constructor attr2
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
						       in ( (SynRule,
								[a1,a2]
							     ),
								
							     (show (i1+1) ++ "_" ++ str1,
							      f ++ " " ++ aux'' cons,
							      [show i2 ++ "_" ++ str2]
							     )
							  ) : r
		HsApp (HsCon cons)
		      (HsVar attr2) -> case expr3 of 
					HsVar attr3 -> -- attr1 = function [Constructor attr2] attr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
							    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
							in  ( (SynRule,
								[a1,a2,a3]
							      ),
							      (show (i1+1) ++ "_" ++ str1,
							       "(\\x y -> " ++ f ++ " (" ++ aux'' cons ++ " x) y)",
							      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3])
							    ): r


-- attr1 = function (expr2) expr3 expr4
patbinds2attrs ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       (HsParen expr2))
							expr3)
						 expr4))
			  y):t) varnames
	= patbinds2attrs ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  y):t) varnames


-- attr1 = function expr2 (expr3) expr4
patbinds2attrs ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							(HsParen expr3))
						 expr4))
			  y):t) varnames
	= patbinds2attrs ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  y):t) varnames


-- attr1 = function expr2 expr3 (expr4)
patbinds2attrs ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 (HsParen expr4)))
			  y):t) varnames
	= patbinds2attrs ((HsPatBind x
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       expr2)
							expr3)
						 expr4))
			  y):t) varnames


-- attr1 = function attr2 attr3 attr4
patbinds2attrs ((HsPatBind _
			  (HsPVar (HsIdent attr1))
			  (HsUnGuardedRhs (HsApp (HsApp (HsApp (HsVar (UnQual (HsIdent f)))
							       (HsVar attr2))
							(HsVar attr3))
						 (HsVar attr4)))
			  _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	  in ( (SynRule,
		     [a1,a2,a3,a4]
		    ),

		    (show (i1+1) ++ "_" ++ str1,
		     f,
		     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
		    )
		  )
		:patbinds2attrs t varnames


-- attr1 = function (attr2, attr3)
-- attr1 = function (attr2, attr3, attr4)
patbinds2attrs ((HsPatBind _
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsApp (HsVar (UnQual (HsIdent f)))
						  (HsTuple l)))
			   _):t) varnames 
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      r              = patbinds2attrs t varnames
	  in  case l of
		[HsApp (HsCon (UnQual (HsIdent cons)))
		       (HsTuple [HsVar attr2, HsVar attr3]), HsVar attr4]
					   ->
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
			in ( (SynRule,
				 [a1,a2,a3,a4]
				),
			     (show (i1+1) ++ "_" ++ str1,
 			      "(\\x y z -> " ++ f ++ " (" ++ cons ++ "(x,y), z))",
			      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
			     )
			   ) : r
		[HsVar attr2, HsVar attr3] ->
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
			in ( (SynRule,
				 [a1,a2,a3]
				),
			     ( show (i1+1) ++ "_" ++ str1,
			      "(\\x y -> " ++ f ++ "(x,y))",
			      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
				)
			    ) : r
		[HsVar attr2, HsVar attr3, HsVar attr4] ->
			let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      		    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      		    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
			in ( (SynRule,
				[a1,a2,a3,a4]
				),
			      ( show (i1+1) ++ "_" ++ str1,
				"(\\x y z -> " ++ f ++ "(x,y,z))",
				[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
				)
			   ) : r

	
-- attr1 = (expr1) `op` expr2
patbinds2attrs ((HsPatBind b
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsInfixApp (HsParen expr2)
						       op
						       expr3))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b 
				    (HsPVar (HsIdent attr1))
				    (HsUnGuardedRhs (HsInfixApp expr2
								op
								expr3))
				     x):t) varnames


-- attr1 = expr1 `op` (expr2)
patbinds2attrs ((HsPatBind b
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsInfixApp expr2
						       op
						       (HsParen expr3)))
			   x):t) varnames
	= patbinds2attrs ((HsPatBind b 
				    (HsPVar (HsIdent attr1))
				    (HsUnGuardedRhs (HsInfixApp expr2
								op
								expr3))
				     x):t) varnames

-- attr1 =  attr2  `op`  attr3
-- attr1 = [attr2  `op`  attr3]  `op`  attr4
-- attr1 =  attr2  `op`  literal
-- attr1 = literal `op`  attr2
-- attr1 = [attr2  `op1` attr3]  `op2` literal
-- attr1 = [function attr2]      `op`  literal
-- attr1 = [function attr2]      `op`  attr3
-- attr1 = [attr2  `op`  attr3]  `op` [attr4   `op` attr5]
-- attr1 = [(function attr2)     `op` literal] `op` attr3
-- attr1 = ((f (Cons attr2)) attr3 attr4) `op` attr5
-- attr1 = (Cons (Cons attr2) attr3) `op` attr4
patbinds2attrs ((HsPatBind b
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsInfixApp expr2
						       op
						       expr3))
			   x):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
              r              = patbinds2attrs t varnames
	  in case expr2 of
		HsApp (HsCon (UnQual (HsIdent cons)))
		      (HsTuple [HsVar attr2, HsVar attr3])
			-> case expr3 of
					HsVar attr4 ->
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
							in ( (SynRule,
							      [a1,a2,a3,a4]
							      ),

							    (show (i1+1) ++ "_" ++ str1,
							     "(\\x y z -> (" ++ aux' op ++ ") (" ++ cons ++ " (x,y)) z)",
							     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
								)
							    ):r
		HsApp (HsApp  (HsApp (HsVar f)
			     	     (HsParen (HsApp (HsCon (UnQual (HsIdent cons)))
					     	     (HsVar attr2))))
		              (HsVar attr3))
		      (HsVar  attr4)
			-> case expr3 of 
					HsVar attr5 -> -- attr1 = ((f (Cons attr2)) attr3 attr4) `op` attr5 
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
	      						    a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
							in ( (SynRule,
							      [a1,a2,a3,a4,a5]
								),
							     (show (i1+1) ++ "_" ++ str1,
							     "(\\x y z w -> (" ++ aux' op ++ ") " ++ "((" ++ getVarName f ++ " (" ++ cons ++ " x)) y z) w)",
							     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5]
	       						     )
							   ):r
		HsApp (HsApp (HsCon (UnQual (HsIdent cons1))) 
			     (HsParen (HsApp (HsCon (UnQual (HsIdent cons2)))
			                     (HsVar attr2))))
		      (HsVar attr3)
			-> case expr3 of 
					(HsVar attr4) -> -- attr1 = (Cons (Cons attr2) attr3) `op` attr4
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						    a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
							in ( (SynRule,
							      [a1,a2,a3,a4]
								),
							     (show (i1+1) ++ "_" ++ str1,
								"(\\x y z -> (" ++ aux' op ++ ") (" ++ cons1 ++ " (" ++ cons2 ++" x) y) z)",
								[show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
							      )):r
		HsVar attr2 
			-> case expr3 of
					HsVar attr3 -> -- attr1 = attr2 `op` attr3
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	  					       in ( (SynRule,
							     [a1,a2,a3]
	       						    ),

	       						    (show (i1+1) ++ "_" ++ str1,
							     "(" ++ aux' op ++ ")",
							    [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
	       						    )
	     						  ) : r
					HsLit lit -> -- attr1 = attr2 `op` literal
						     let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
						     in ( (SynRule,
							   [a1,a2]
	       						   ),

	       						  (show (i1+1) ++ "_" ++ str1,
							   "(" ++ aux' op ++ aux lit ++ ")",
							   [show i2 ++ "_" ++ str2]
	       						  )
	     						) : r
		HsInfixApp (HsVar attr2)
			   op1
			   (HsVar attr3) 
			-> case expr3 of
					HsVar attr4 -> -- attr1 = [attr2 `op` attr3] `op` attr4
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						   a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
						       in ( (SynRule,
								[a1,a2,a3,a4]
	       						    ),

	       						    (show (i1+1) ++ "_" ++ str1,
							    "(\\x y z -> ((x " ++ aux' op1 ++ " y) " ++ aux' op ++ " z))",
					 		    [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4]
       							    )
	     						  ) : r
					HsLit lit   -> -- attr1 = [attr2 `op1` attr3] `op2` literal
						        let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
							in ( (SynRule,
							      [a1,a2,a3]
	       						     ),

	       						     (show (i1+1) ++ "_" ++ str1,
							      "(\\x y -> ((x " ++ aux' op1 ++ " y) " ++ aux' op ++ " " ++ aux lit ++ "))",
							      [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
	       						     )
	     						   ) : r
					HsInfixApp (HsVar attr4)
						   op3
						   (HsVar attr5)
						    -> -- attr1 = [attr2 `op` attr3] `op` [attr4 `op` attr5]
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
	      						   a4@(_,i4,str4) = eficientPartition isDigit varnames (getVarName attr4)
						           a5@(_,i5,str5) = eficientPartition isDigit varnames (getVarName attr5)
	  						in ( (SynRule,
							      [a1,a2,a3,a4,a5]
	       						     ),

	       						     (show (i1+1) ++ "_" ++ str1,
							     "(\\x y z w-> ((x " ++ aux' op1 ++ " y) " ++ aux' op ++ "(z" ++ aux' op3 ++ "w)))",
							     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3, show i4 ++ "_" ++ str4, show i5 ++ "_" ++ str5]
	       						     )
	     						   ) : r
		HsLit lit
			-> case expr3 of 
					HsVar attr3 -> -- attr1 = literal `op` attr2
						       let a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
						       in ( (SynRule,
							     [a1,a3]
							     ),

							    (show (i1+1) ++ "_" ++ str1,
		 					    aux lit ++ aux' op,
		 					    [show i3 ++ "_" ++ str3]
							    )
	     						  ) : r
		HsApp  (HsVar f)
		       (HsVar attr2)
			-> case expr3 of
					HsLit lit   -> -- attr1 = [function attr2] `op` literal
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
						       in ( (SynRule,
							     [a1,a2]
	       						     ),

							     (show (i1+1) ++ "_" ++ str1,
						 	     "(\\ x -> (" ++ FunctionTypes.space (getVarName f) ++ "x) " ++ FunctionTypes.space (aux' op) ++ aux lit ++ ")",
							     [show i2 ++ "_" ++ str2])
	     						  ) : r
					HsVar attr3 -> -- attr1 = [function attr2] `op` attr3
							let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						    a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
							in ( (SynRule,
							      [a1,a2,a3]
	       						     ),

	       						     (show (i1+1) ++ "_" ++ str1,
							     "(\\x y -> (" ++ FunctionTypes.space (getVarName f) ++ "x) " ++ FunctionTypes.space (aux' op) ++ "y)",
	       						     [show i2 ++ "_" ++ str2, show i3 ++ "_" ++ str3]
							     )	
	     						   ) : r
		HsInfixApp (HsParen (HsApp (HsVar f)
					   (HsVar attr2)))
			   op1
			   (HsLit lit)
			-> case expr3 of 
					HsVar attr3 -> -- attr1 = [(function attr2) `op` literal] `op` attr3
						       let a2@(_,i2,str2) = eficientPartition isDigit varnames (getVarName attr2)
	      						   a3@(_,i3,str3) = eficientPartition isDigit varnames (getVarName attr3)
						       in ( (SynRule,
							     [a1,a2,a3]
	       						     ),

							    (show (i1+1) ++ "_"++ str1,
		 					     "(\\x y -> ((" ++ FunctionTypes.space (getVarName f) ++  "x) " 
			  				     ++ FunctionTypes.space (aux' op1) ++ aux lit ++ ")" ++ aux' op ++ " y)",
		 					     [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3]
	        					    )
	     						 ) : r


-- attr1 = if (expr1) then expr2 else expr3
patbinds2attrs ((HsPatBind x
			   attr1
			   (HsUnGuardedRhs (HsIf (HsParen expr1) expr2 expr3)) y):t) varnames
	= patbinds2attrs ((HsPatBind x attr1 (HsUnGuardedRhs (HsIf expr1 expr2 expr3)) y):t) varnames


-- attr1 = if expr1 then (expr2) else expr3
patbinds2attrs ((HsPatBind x
			   attr1
			   (HsUnGuardedRhs (HsIf expr1 
						(HsParen expr2) 
						expr3)) 
			   y):t) varnames
	= patbinds2attrs ((HsPatBind x attr1 (HsUnGuardedRhs (HsIf expr1 expr2 expr3)) y):t) varnames


-- attr1 = if expr1 then expr2 else (expr3)
patbinds2attrs ((HsPatBind x
			   attr1
			   (HsUnGuardedRhs (HsIf expr1 
						 expr2 
						 (HsParen expr3))) 
			   y):t) varnames
	= patbinds2attrs ((HsPatBind x attr1 (HsUnGuardedRhs (HsIf expr1 expr2 expr3)) y):t) varnames


-- attr1 = if [b attr2] then [attr3] else [attr4]
patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b)))  (HsVar (UnQual (HsIdent attr2))))
						 (HsVar (UnQual (HsIdent attr3)))
						 (HsVar (UnQual (HsIdent attr4)))))
			    y):t) varnames
	= patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b)))  (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr4))))))
			   y):t) varnames


-- attr1 = if [b attr2] then [f attr3] else [attr4]
patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr3))))
						 (HsVar (UnQual (HsIdent attr4)))))
			    y):t) varnames
	= patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr4))))))
			   y):t) varnames


-- attr1 = if [b attr2] then [attr3] else [f attr4]
patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsVar (UnQual (HsIdent attr3)))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr4))))))
			    y):t) varnames
	= patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b))) (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent f))) (HsVar (UnQual (HsIdent attr4))))))
			   y):t) varnames


-- attr1 = if [b attr2] then [f1 attr3] else [f2 attr4]
-- assuming that no parenthesis are used to establish priorities
-- the "if ... then ... else ..." is really necessary in order to produce haskell function macros in the way that the cpp accepts them
patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsApp (HsVar (UnQual (HsIdent b)))  (HsVar (UnQual (HsIdent attr2))))
						 (HsApp (HsVar (UnQual (HsIdent f1))) (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent f2))) (HsVar (UnQual (HsIdent attr4))))))
			   _):t) varnames
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
	      a3@(_,i3,str3) = eficientPartition isDigit varnames attr3
	      a4@(_,i4,str4) = eficientPartition isDigit varnames attr4
	in ( (SynRule,
		nub [a1,a2,a3,a4]
	       ),

		(show (i1+1) ++ "_"++ str1,
		 if (attr2==attr3) then "(\\x y -> if (" ++ b  ++  " x) then (" ++ f1 ++ " x) else (" ++ f2 ++" y))"
					       else (if (attr2==attr4) 
							then "(\\x y -> if (" ++ b  ++  " x) then (" ++ f1 ++ " y) else (" ++ f2 ++ " x))" 
							else (if (attr3==attr4) 
								then "(\\x y -> if (" ++ b  ++  " x) then (" ++ f1 ++ " y) else (" ++ f2 ++ " y))"
								else "(\\x y z -> if (" ++ b  ++  " x) then (" ++ f1 ++ " y) else (" ++ f2 ++ " z))")),
		 nub [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4]
	        )
	     )
		:patbinds2attrs t varnames


-- attr1 = if [attr2 `op` attr3] then attr4 else attr5
-- assuming that no parenthesis are used to establish priorities
patbinds2attrs ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							     op
							     (HsVar attr3)) 
						 (HsVar attr4)
						 (HsVar attr5))) y):t) varnames
	= patbinds2attrs ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							               op
							               (HsVar attr3))
						            (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr4))
							    (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr5)))) y):t) varnames


-- attr1 = if [attr2 `op` attr3] then [f attr4] else attr5
-- assuming that no parenthesis are used to establish priorities
patbinds2attrs ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							     op
							     (HsVar attr3)) 
						 (HsApp (HsVar f) (HsVar attr4))
						 (HsVar attr5))) y):t) varnames
	= patbinds2attrs ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							               op
							               (HsVar attr3))
						            (HsApp (HsVar f) (HsVar attr4))
							    (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr5)))) y):t) varnames


-- attr1 = if [attr2 `op` attr3] then [f attr4] else attr5
-- assuming that no parenthesis are used to establish priorities
patbinds2attrs ((HsPatBind x
			   (HsPVar attr1)
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							     op
							     (HsVar attr3)) 
						 (HsVar attr4)
						 (HsApp (HsVar f) (HsVar attr5)))) y):t) varnames
	= patbinds2attrs ((HsPatBind x
				     (HsPVar attr1)
				     (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar attr2)
							               op
							               (HsVar attr3))
						            (HsApp (HsVar (UnQual (HsIdent "id"))) (HsVar attr4))
							    (HsApp (HsVar f) (HsVar attr5)))) y):t) varnames


-- attr1 = if [attr2 `op` attr3] then [f1 attr4] else [f2 attr5]
-- the "if ... then ... else ..." chain is really necessary in order to produce haskell function macros 
-- in the way that the cpp accepts them
patbinds2attrs ((HsPatBind x
			   (HsPVar (HsIdent attr1))
			   (HsUnGuardedRhs (HsIf (HsInfixApp (HsVar (UnQual (HsIdent attr2)))
							     op
							     (HsVar (UnQual (HsIdent attr3))))
						 (HsApp (HsVar (UnQual (HsIdent f1))) (HsVar (UnQual (HsIdent attr4))))
						 (HsApp (HsVar (UnQual (HsIdent f2))) (HsVar (UnQual (HsIdent attr5))))))
			    _):t) varnames 
	= let a1@(_,i1,str1) = eficientPartition isDigit varnames attr1
	      a2@(_,i2,str2) = eficientPartition isDigit varnames attr2
	      a3@(_,i3,str3) = eficientPartition isDigit varnames attr3
	      a4@(_,i4,str4) = eficientPartition isDigit varnames attr4
	      a5@(_,i5,str5) = eficientPartition isDigit varnames attr5
	  in (	(SynRule,
		 nub [a1,a2,a3,a4,a5]
		),
		
		(show (i1+1) ++ "_"++ str1,
		 if (attr2==attr4 && attr3==attr5) then "(\\x y -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " x) else (" ++ f2 ++ " y))"
					       else (if (attr2==attr5 && attr3==attr4) 
							then "(\\x y -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " y) else (" ++ f2 ++ " x))" 
							else (if (attr2==attr5) 
								then "(\\x y z -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " z) else (" ++ f2 ++ " x))"
								else (if (attr3==attr4) 
									then "(\\x y z -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " y) else (" ++ f2 ++ " z))" 
									else (((if (attr2==attr4) 
										 then "(\\x y z   -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " x) else (" ++ f2 ++ " z))"
							        		 else ( if (attr3==attr5) 
											then "(\\x y z   -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " z)  else (" ++ f2 ++ " y))"
											else "(\\x y z w -> if (x " ++ (aux' op)  ++  " y) then (" ++ f1 ++ " z) else (" ++ f2 ++ " w))"))))))),
		 nub [show i2 ++ "_" ++ str2,show i3 ++ "_" ++ str3,show i4 ++ "_" ++ str4,show i5 ++ "_" ++ str5])
	     )
		:patbinds2attrs t varnames


-- 
patbinds2attrs (x:t) 
	       varnames 
	= error (show x) $ patbinds2attrs t varnames 



{- 
   We use 'eficientPartition' for partitioning strings in two: the non-numeric part and the numeric part!
   The function is more general, though, since it is possible to use another criteria for partitioning 
	strings.

   Our strings correspond to attributes and their numeric parts, if any, at the end, correspond to attribute
	occurrences,

   For example, if we have the following grammar production:

   Tree: Fork (Int Tree Tree)

   Then, when using "ilist2", we are referring to the attribut "ilist" of the first rhs occurrence of "Tree" 

-}

eficientPartition :: (Char -> Bool) 
		  -> [String] 
	          -> [Char]
	          -> (T, Int, [Char])
eficientPartition p varnames attr 
	= if (attr `elem` varnames)
	  then (N, (\(Just x) -> x+1) (elemIndex attr  varnames),"")
	  else 
	    let (str,n) = eficientPartition' p attr
	    in if (n == [])
	       then (A, 0,                           str)
	       else (A, ((read :: String -> Int) n), str)
	where eficientPartition' _ [] = ([],[])
	      eficientPartition' p (h:t) = let (ac1',ac2') = eficientPartition' p t 
					   in if (p h) 
					      then (ac1',h:ac2')
					      else (h:ac1',ac2')

{- 
'getVarName' is just an auxiliary function to acess a Haskell variable name and to avoid writing 
	'(UnQual (HsIdent ...))' everywhere in the code
-}
getVarName (UnQual (HsIdent name)) = name


aux (HsInt int)       = show int
aux (HsChar char)     = show char
aux (HsString string) = show string
aux (HsFrac frac)     = show frac
aux' (HsQVarOp name)  = aux'' name
aux' (HsQConOp name)  = aux'' name
aux'' (Qual _ (HsIdent name))  = name
aux'' (Qual _ (HsSymbol name)) = name
aux'' (UnQual (HsIdent name))  = name
aux'' (UnQual (HsSymbol name)) = name
aux'' (Special HsCons)         = ":"


